"""
This module contains the data pipeline container for organising and 
calcuating the required summary statistics
"""

# Authors: Joseph Moorhouse 
#
# License: BSD 3 clause

import pandas as pd 

# --------------------------------------------------------------------------
# Data pipeline

class SalesPipeline:
    """Backend pipeline for data in sales_analysis/data_pipeline/data
    
    Parameters
    ----------
    data : dict of pd.DataFrame
        dict should contain sales data from 
        sales_analysis/data_pipeline/data. Must include 'commissions.csv', 
        'orders.csv', 'order_lines.csv', 'products.csv', 
        'product_promotions.csv' and 'promotions.csv'

    Attributes
    ----------
    merged_order_data : pd.DataFrame
        Formatted order/orderlines data
    """
    # ----------------------------------------------------------------------
    # Constructors

    def __init__(self, **data):
        for k, v in data.items():
            setattr(self, k.split('.')[0], v)
            
        self.merged_order_data = self._merge_order_data()

    # ----------------------------------------------------------------------
    # formatting methods 
            
    def _format_orders(self, orders):
        "date formatting method for orders.csv"

        orders["created_at"] = pd.to_datetime(orders["created_at"]).dt.date
        return orders
    
    def _merge_order_data(self):
        "Method to combine orders.csv and order_lines.csv"

        orders = self._format_orders(self.orders)
        order_lines = self.order_lines
        
        return order_lines.merge(
            orders, 
            how="left", 
            left_on="order_id", 
            right_on="id",
        )

    # ----------------------------------------------------------------------
    # statistics/calculation methods

    def _customer_count(self):
        "Method to count the total number of unique customers per day"

        orders = self._format_orders(self.orders)
        customer_count = orders.groupby("created_at")["customer_id"].nunique()
        customer_count.name = "customers"
        
        return customer_count
    
    def _total_discount(self):
        "Method to calculate the total discount given per day"

        merged = self.merged_order_data  
        merged["total_discount"] = (
            merged["full_price_amount"] - merged["discounted_amount"])
        
        total_discount = merged.groupby("created_at")["total_discount"].sum()
        total_discount.name = "total_discount_amount"
        
        return total_discount
        
    def _item_count(self):
        "Method to calculate the total number of items sold per day"

        sales_quantity = (self.merged_order_data
                 .groupby("created_at")["quantity"].sum())
        sales_quantity.name = "items"
        
        return sales_quantity
    
    def _mean_order_total(self):
        "Method to calculate the average order amount per day"

        average_daily_order = (self.merged_order_data
                .groupby("created_at")['total_amount'].mean())
        average_daily_order.name = "order_total_avg"
        
        return average_daily_order
    
    def _mean_discount_rate(self):
        "Method to calculate the average discount rate offered per day"

        average_discount_rate = (self.merged_order_data
                .groupby("created_at")["discount_rate"].mean())
        average_discount_rate.name = "discount_rate_avg"
        
        return average_discount_rate

    # ----------------------------------------------------------------------
    # Public methods
    
    def summary(self):  
        """Summary sales statistics per day.
        
        Returns
        -------
        summary: pd.DataFrame
            Summary of the total number of customers, total discount offered,
            total number of items sold, average order amount and the average 
            discount rate offered each day
        """

        summary_df = pd.concat([
            self._customer_count(),
            self._total_discount(),
            self._item_count(),
            self._mean_order_total(),
            self._mean_discount_rate(),
        ], axis=1)
        
        summary_df.index = pd.to_datetime(summary_df.index)
        return summary_df


import pytest
from sales_analysis.application.app import sales_app

@pytest.fixture
def app():
    return sales_app

def test_root(client):
    response = client.get("/")
    assert response.status_code == 404

@pytest.mark.parametrize(
    "date, solution", 
    [("2019-08-01", {"customers":9.0,
                     "discount_rate_avg":0.1252497888814673,
                     "items":2895.0,
                     "order_total_avg":1182286.0960463749,
                     "total_discount_amount":15152814.736907512}),
     ("2019-09-29", {"customers":5.0,
                     "discount_rate_avg":0.17648554006271688,
                     "items":1544.0,
                     "order_total_avg":993268.9252802497,
                     "total_discount_amount":12999485.945880124}),
     ("2019-09-01", {"customers":8.0,
                     "discount_rate_avg":0.15735061266940567,
                     "items":2699.0,
                     "order_total_avg":1184520.138455093,
                     "total_discount_amount":18164159.9892837}),
    ]
)
def test_dates(client, date, solution):
    response = client.get(f"/{date}")
    assert response.get_json() == solution

@pytest.mark.parametrize(
    "date", [("019-08-01"), ("2019-090-29"), ("2019-09-010"), ("a")])
def test_bad_dates(client, date):
    response = client.get(f"/{date}")
    assert str(response.get_data(), "utf-8") == (
        "Invalid format. Please follow %Y-%m-%d pattern")

@pytest.mark.parametrize(
    "date", [("1900-08-01"), ("2100-09-01")])
def test_missing_date(client, date):
    response = client.get(f"/{date}")
    assert str(response.get_data(), "utf-8") == (
        f"No data for {date}. Please try another date, such as 2019-08-01")


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

def setup_package():
    setup(
        name='sales-analysis',
        version='0.4',
        author='Joseph Moorhouse',
        license=license,
        description='A sales analysis package for an imaginary e-shop',
        long_description_content_type="text/markdown",
        long_description=long_description,
        packages=find_packages(include=["sales_analysis", "sales_analysis.*"]),
        include_package_data = True,
        url="https://github.com/jbmoorhouse/sales_analysis",
        install_requires=[
            "flask>=1.1.1",
            "pytest>=5.2",
            "pandas>=1.0.0",
            "pytest-flask>=1.0.0"
        ],
        classifiers=[
            "Programming Language :: Python :: 3.7",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.7',
    )

if __name__ == "__main__": 
    setup_package()


"""
This module contains the web application to visualise the summary data
from the data pipeline
"""

# Authors: Joseph Moorhouse 
#
# License: BSD 3 clause

from flask import Flask
from datetime import datetime
import os
import pandas as pd

from sales_analysis.data_pipeline._pipeline import SalesPipeline
from sales_analysis.data_pipeline import BASEPATH

# --------------------------------------------------------------------------
# Load and check valid data

FILEPATH = os.path.join(BASEPATH, "data")
DATA_FILES = [f for f in os.listdir(FILEPATH) if f.endswith('.csv')]
REQUIRED_FILES = [
    'commissions.csv', 
    'orders.csv', 
    'order_lines.csv', 
    'products.csv', 
    'product_promotions.csv', 
    'promotions.csv'
]

for file in REQUIRED_FILES:
    if os.path.basename(file) not in DATA_FILES:
        raise FileNotFoundError(f"{os.path.basename(file)} was not found")

DATA = {f : pd.read_csv(os.path.join(FILEPATH, f)) for f in DATA_FILES}

# --------------------------------------------------------------------------
# Web application

sales_app = Flask("sales_analysis")

@sales_app.route('/<selected_date>', methods=['GET'])
def my_view(selected_date):
    try:
        date = datetime.strptime(selected_date, "%Y-%m-%d").date()
    except ValueError:
        return "Invalid format. Please follow %Y-%m-%d pattern"

    sales = SalesPipeline(**DATA)
    sales_summary = sales.summary()

    try:
        daily_data = sales_summary.loc[date]
        return daily_data.to_dict()
    except (KeyError):
        return f"No data for {date}. Please try another date, " \
                "such as 2019-08-01"""

if __name__ == "__main__":
    sales_app.run(debug=False)